# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午2:38
@file: basicblock.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn

from zcls.model.attention_helper import make_attention_block


class BasicBlock(nn.Module, ABC):
    expansion = 1

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride=1,
                 downsample=None,
                 groups=1,
                 base_width=64,
                 with_attention=False,
                 reduction=16,
                 attention_type='SqueezeAndExcitationBlock2D',
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 **kwargs
                 ):
        """
        使用两个3x3卷积，如果进行下采样，那么使用第一个卷积层对输入空间尺寸进行减半操作
        参考Torchvision实现
        对于注意力模块，有两种嵌入方式：
        1. 对于Squeeze-And-Excitation或者Global Context操作，在残差连接中（after 1x1）嵌入；
        2. 对于NonLocal或者SimplifiedNonLoal，在Block完成计算后（after add）嵌入。
        :param in_channels:　输入通道数
        :param out_channels:　输出通道数
        :param stride:　步长
        :param downsample:　下采样
        :param groups:　cardinality
        :param base_width:　基础宽度
        :param with_attention:　是否使用注意力模块
        :param reduction:　衰减率
        :param attention_type:　注意力模块类型
        :param conv_layer:　卷积层类型
        :param norm_layer:　归一化层类型
        :param act_layer:　激活层类型
        :param kwargs:　其他参数
        """
        super(BasicBlock, self).__init__()
        assert with_attention in (0, 1)
        assert attention_type in ['GlobalContextBlock2D',
                                  'SimplifiedNonLocal2DEmbeddedGaussian',
                                  'NonLocal2DEmbeddedGaussian',
                                  'SqueezeAndExcitationBlock2D']
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.down_sample = downsample

        self.conv1 = conv_layer(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = norm_layer(out_channels)

        self.conv2 = conv_layer(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = norm_layer(out_channels)

        self.relu = act_layer(inplace=True)

        self.attention_after_1x1 = None
        self.attention_after_add = None
        if with_attention and attention_type in ['SqueezeAndExcitationBlock2D', 'GlobalContextBlock2D']:
            self.attention_after_1x1 = make_attention_block(out_channels * self.expansion, reduction, attention_type)
            self.attention_after_add = None
        if with_attention and attention_type in ['NonLocal2DEmbeddedGaussian', 'SimplifiedNonLocal2DEmbeddedGaussian']:
            self.attention_after_1x1 = None
            self.attention_after_add = make_attention_block(out_channels * self.expansion, reduction, attention_type)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.attention_after_1x1 is not None:
            out = self.attention_after_1x1(out)

        if self.down_sample is not None:
            identity = self.down_sample(x)

        out += identity
        out = self.relu(out)

        if self.attention_after_add is not None:
            out = self.attention_after_add(out)

        return out



# -*- coding: utf-8 -*-

"""
@date: 2021/6/3 下午8:26
@file: ghost_bottleneck.py
@author: zj
@description: 
"""

import torch.nn as nn

from zcls.model.attention_helper import make_attention_block
from .ghost_module import GhostModule


class GhostBottleneck(nn.Module):

    def __init__(self,
                 in_channels,
                 mid_channels,
                 out_channels,
                 stride=1,
                 kernel_size=3,
                 with_attention=False,
                 reduction=4,
                 attention_type='SqueezeAndExcitationBlock2D'
                 ):
        """
        when stride=1, bottleneck=Add(
                                    Input,
                                    BN(GhostModule(
                                        ReLU(BN(
                                            GhostModule(Input)
                                        ))
                                    ))
                                    )
        when stride=2, bottleneck=Add(
                                    Input,
                                    BN(GhostModule(
                                        BN(DWConv(
                                            ReLU(BN(
                                                GhostModule(Input)
                                            ))
                                        ))
                                    ))
                                    )
        if assert squeeze-and-excitation module, use it before ghost2
        :param in_channels: 输入通道数
        :param mid_channels: 膨胀通道数
        :param out_channels: 输出通道数
        :param stride: 步长，是否执行下采样操作
        :param kernel_size: 深度卷积核大小
        :param with_attention: 是否执行注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力类型
        """
        super(GhostBottleneck, self).__init__()

        # Point-wise expansion
        self.ghost1 = GhostModule(in_channels, mid_channels, is_act=True)

        self.stride = stride
        # Depth-wise convolution
        if self.stride > 1:
            padding = (kernel_size - 1) // 2
            self.conv_dw = nn.Conv2d(mid_channels, mid_channels, (kernel_size, kernel_size), stride=(stride, stride),
                                     padding=(padding, padding), groups=mid_channels, bias=False)
            self.bn_dw = nn.BatchNorm2d(mid_channels)

        # Squeeze-and-excitation
        if with_attention:
            self.se = make_attention_block(mid_channels,
                                           reduction,
                                           attention_type,
                                           sigmoid_type='HSigmoid',
                                           is_round=True,
                                           round_nearest=4,
                                           bias=True,
                                           )
        self.with_attention = with_attention

        # Point-wise linear projection
        self.ghost2 = GhostModule(mid_channels, out_channels, is_act=False)

        # shortcut
        if in_channels == out_channels and self.stride == 1:
            self.shortcut = nn.Sequential()
        else:
            padding = (kernel_size - 1) // 2
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_channels, in_channels, (kernel_size, kernel_size), stride=(stride, stride),
                          padding=(padding, padding), groups=in_channels, bias=False),
                nn.BatchNorm2d(in_channels),
                nn.Conv2d(in_channels, out_channels, (1, 1), stride=(1, 1), padding=(0, 0), bias=False),
                nn.BatchNorm2d(out_channels),
            )

    def forward(self, x):
        residual = x

        # 1st ghost bottleneck
        x = self.ghost1(x)

        # Depth-wise convolution
        if self.stride > 1:
            x = self.conv_dw(x)
            x = self.bn_dw(x)

        # Squeeze-and-excitation
        if self.with_attention:
            x = self.se(x)

        # 2nd ghost bottleneck
        x = self.ghost2(x)

        x += self.shortcut(residual)
        return x



# -*- coding: utf-8 -*-

"""
@date: 2020/12/21 下午8:17
@file: label_smoothing_loss.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn
import torch.nn.functional as F

from .. import registry
from zcls.config.key_word import KEY_OUTPUT, KEY_LOSS


@registry.CRITERION.register('LabelSmoothingLoss')
class LabelSmoothingLoss(nn.Module, ABC):
    """
    Label smoothing cross entropy loss
    Refer to:
    1. [解决过拟合：如何在PyTorch中使用标签平滑正则化](https://zhuanlan.zhihu.com/p/123077402)
    2. [Rethinking the Inception Architecture for Computer Vision](https://arxiv.org/abs/1512.00567)
    3. [[译]Rethinking the Inception Architecture for Computer Vision](https://blog.zhujian.life/posts/a0a2be91.html)
    """

    def __init__(self, cfg):
        super(LabelSmoothingLoss, self).__init__()
        self.epsilon = 0.1
        self.reduction = 'mean'

    def __call__(self, output_dict, targets):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()
        inputs = output_dict[KEY_OUTPUT]

        n = inputs.size()[-1]
        log_preds = F.log_softmax(inputs, dim=-1)
        loss = self.reduce_loss(-log_preds.sum(dim=-1), reduction=self.reduction)
        nll = F.nll_loss(log_preds, targets, reduction=self.reduction)

        return {KEY_LOSS: self.linear_combination(loss / n, nll, self.epsilon)}

    def reduce_loss(self, loss, reduction='mean'):
        return loss.mean() if reduction == 'mean' else loss.sum() if reduction == 'sum' else loss

    def linear_combination(self, x, y, epsilon):
        return epsilon * x + (1 - epsilon) * y



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:51
@file: transform.py
@author: zj
@description: 
"""

from yacs.config import CfgNode as CN
from ztransforms.cls.autoaugment import AutoAugmentPolicy


def add_config(_C):
    # ---------------------------------------------------------------------------- #
    # Transform
    # ---------------------------------------------------------------------------- #
    _C.TRANSFORM = CN()
    _C.TRANSFORM.TRAIN_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')
    _C.TRANSFORM.TEST_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')

    # If size is a sequence like (h, w), output size will be matched to this.
    # If size is an int, smaller edge of the image will be matched to this number.
    # i.e, if height > width, then image will be rescaled to (size * height / width, size).
    _C.TRANSFORM.TRAIN_RESIZE = (224, )
    _C.TRANSFORM.TEST_RESIZE = (224, )

    # Desired output size of the crop.
    # If size is an int instead of sequence like (h, w), a square crop (size, size) is made.
    # If provided a sequence of length 1, it will be interpreted as (size[0], size[0]).
    _C.TRANSFORM.TRAIN_CROP = (224, 224)
    _C.TRANSFORM.TEST_CROP = (224, 224)

    # brightness (float or tuple of float (min, max)): How much to jitter brightness.
    #     brightness_factor is chosen uniformly from [max(0, 1 - brightness), 1 + brightness]
    #     or the given [min, max]. Should be non negative numbers.
    # contrast (float or tuple of float (min, max)): How much to jitter contrast.
    #     contrast_factor is chosen uniformly from [max(0, 1 - contrast), 1 + contrast]
    #     or the given [min, max]. Should be non negative numbers.
    # saturation (float or tuple of float (min, max)): How much to jitter saturation.
    #     saturation_factor is chosen uniformly from [max(0, 1 - saturation), 1 + saturation]
    #     or the given [min, max]. Should be non negative numbers.
    # hue (float or tuple of float (min, max)): How much to jitter hue.
    #     hue_factor is chosen uniformly from [-hue, hue] or the given [min, max].
    #     Should have 0<= hue <= 0.5 or -0.5 <= min <= max <= 0.5.
    # (brightness, contrast, saturation, hue)
    _C.TRANSFORM.ColorJitter = (0.1, 0.1, 0.1, 0.1)

    # AutoAugment policies learned on different datasets.
    _C.TRANSFORM.AUGMENT_POLICY = "imagenet"

    # for AutoAugment, only torch.uint8 image tensors are supported
    # for Normalize, may be happens
    # ValueError: std evaluated to zero after conversion to torch.uint8, leading to division by zero.
    _C.TRANSFORM.IMAGE_DTYPE = 'float32'

    # Sequence of means for each channel.
    _C.TRANSFORM.MEAN = (0.45, 0.45, 0.45)

    # Sequence of standard deviations for each channel.
    _C.TRANSFORM.STD = (0.225, 0.225, 0.225)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/2 下午9:46
@file: mobilenetv1_block.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn


class MobileNetV1Block(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride=1,
                 padding=1,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ) -> None:
        """
        Block = depth-wise convolution + point-wise convolution
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param stride: 卷积层步长
        :param padding: 卷积层零填充
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV1Block, self).__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.conv1 = conv_layer(in_channels, in_channels, kernel_size=3, stride=stride, padding=padding,
                                groups=in_channels, bias=False)
        self.bn1 = norm_layer(in_channels)
        self.relu = act_layer(inplace=True)

        self.conv2 = conv_layer(in_channels, out_channels, kernel_size=1, stride=1, bias=False)
        self.bn2 = norm_layer(out_channels)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)

        return x



# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午8:53
@file: acb_util.py
@author: zj
@description: 
"""

import torch


def _fuse_kernel(kernel, gamma, std):
    b_gamma = torch.reshape(gamma, (kernel.shape[0], 1, 1, 1))
    b_gamma = b_gamma.repeat(1, kernel.shape[1], kernel.shape[2], kernel.shape[3])
    b_std = torch.reshape(std, (kernel.shape[0], 1, 1, 1))
    b_std = b_std.repeat(1, kernel.shape[1], kernel.shape[2], kernel.shape[3])
    return kernel * b_gamma / b_std


def _add_to_square_kernel(square_kernel, asym_kernel):
    asym_h = asym_kernel.shape[2]
    asym_w = asym_kernel.shape[3]
    square_h = square_kernel.shape[2]
    square_w = square_kernel.shape[3]
    square_kernel[:, :, square_h // 2 - asym_h // 2: square_h // 2 - asym_h // 2 + asym_h,
    square_w // 2 - asym_w // 2: square_w // 2 - asym_w // 2 + asym_w] += asym_kernel
